# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2021


from sanic.views import HTTPMethodView
from sanic.response import text


class SimpleView(HTTPMethodView):
    def get(self, request):
        return text("I am get method")

    # You can also use async syntax
    async def post(self, request):
        return text("I am post method")

    def put(self, request):
        return text("I am put method")

    def patch(self, request):
        return text("I am patch method")

    def delete(self, request):
        return text("I am delete method")
